#!/home/rob/.env/topics/bin/python
import math
import random
from ..wrappers import dLDAMallet, deTNDMallet
from .dnlda import dNLDA
from ..helpers.common import save_topics


class deNLDA(dNLDA):
    topics = []

    def __init__(self, dataset=None, k=30, tnd_alpha=50, tnd_beta0=0.01, tnd_beta1=25, tnd_noise_words_max=200,
                 tnd_iterations=1000, lda_iterations=1000, lda_alpha=50, lda_beta=0.01, nlda_phi=10,
                 nlda_topic_depth=100, top_words=20, nlda_phi_list=None, nlda_topic_depth_list=None,
                 num_time_periods=10, tnd_noise_distribution=[], lda_tw_dist=[], lda_topics=[], corpus=None,
                 dictionary=None, last_tnd_alpha_array_file=None, last_tnd_noise_dist_file=None,
                 last_tnd_tw_dist_file=None, last_lda_alpha_array_file=None, last_lda_tw_dist_file=None,
                 lda_save_path=None, save_path=None, mallet_tnd_path=None, mallet_lda_path=None, embedding_path=None,
                 closest_x_words=3, random_seed=1824, workers=4, run=True):

        super().__init__(dataset, k, tnd_alpha, tnd_beta0, tnd_beta1, tnd_noise_words_max, tnd_iterations, lda_iterations,
                       lda_alpha, lda_beta, nlda_phi, nlda_topic_depth, top_words, nlda_phi_list, nlda_topic_depth_list,
                       num_time_periods, tnd_noise_distribution, lda_tw_dist, lda_topics, corpus, dictionary,
                       last_tnd_alpha_array_file, last_tnd_noise_dist_file, last_tnd_tw_dist_file, last_lda_alpha_array_file,
                       last_lda_tw_dist_file, lda_save_path, save_path, mallet_tnd_path, mallet_lda_path, random_seed,
                       workers, run=False)
        if save_path is not None:
            save = True
            self.save_path = save_path
        else:
            save = False
            self.save_path = 'denlda_results/'
        self.embedding_path = embedding_path
        self.closest_x_words = closest_x_words

        if run:
            self.run_nlda_all_time_periods(save=save)

    def compute_nlda(self, noise_dist, tw_dist, lda_topics, phi, depth):
        """
        takes self.tnd_noise_distribution, self.lda_tw_dist, self.phi, self.top_words, and computes NLDA topics
        sets self.topics to the set of topics computed from noise distribution and topic word distribution
        :return: void
        """
        topics = []
        for i in range(0, len(lda_topics)):
            topic = lda_topics[i]
            final_topic = []
            j = 0
            while len(topic) > j and len(final_topic) < self.top_words and j < depth:
                w = topic[j]
                token_id = self.dictionary.token2id[w]
                beta = 2
                if w in noise_dist:
                    beta += noise_dist[w]
                beta = max(2, beta * (phi / self.k))
                alpha = 2 + tw_dist[i, token_id]
                roll = random.betavariate(alpha=math.sqrt(alpha), beta=math.sqrt(beta))
                if roll >= 0.5:
                    final_topic.append(w)
                    if w not in noise_dist:
                        noise_dist[w] = 0
                    noise_dist[w] += (alpha - 2)
                j += 1
            topics.append(final_topic)
        return topics

    def run_nlda_one_time_period(self, t):
        self.prepare_data(t)
        if len(self.tnd_noise_distribution) > t:
            noise_dist = self.tnd_noise_distribution[t]
        else:
            tnd_model = deTNDMallet(self.mallet_tnd_path, corpus=self.corpus, num_topics=self.k, beta=self.last_tnd_beta,
                                    id2word=self.dictionary, iterations=self.tnd_iterations, skew=self.tnd_beta1,
                                    noise_words_max=self.tnd_noise_words_max, workers=self.tnd_workers,
                                    noise_dist_file=self.last_tnd_noise_dist_file,
                                    tw_dist_file=self.last_tnd_tw_dist_file,
                                    alpha_array_infile=self.last_tnd_alpha_array_file, alpha=self.tnd_alpha,
                                    embedding_path=self.embedding_path, closest_x_words=self.closest_x_words)
            noise_dist = tnd_model.load_noise_dist()
            self.tnd_noise_distribution.append(noise_dist)
            self.last_tnd_beta = tnd_model.load_beta()
            self.last_tnd_alpha_array_file = tnd_model.falphaarrayfile()
            self.last_tnd_noise_dist_file = tnd_model.fnoisefile()
            self.last_tnd_tw_dist_file = tnd_model.fwordweights()

        if len(self.lda_tw_dist) > t and len(self.lda_topics) > t:
            tw_dist = self.lda_tw_dist[t]
            lda_topics = self.lda_topics[t]
        else:
            lda_model = dLDAMallet(self.mallet_lda_path, self.corpus, num_topics=self.k, id2word=self.dictionary,
                               workers=self.lda_workers, tw_dist_file=self.last_lda_tw_dist_file,
                               alpha_array_infile=self.last_lda_alpha_array_file, beta=self.last_lda_beta,
                               iterations=self.lda_iterations, random_seed=self.random_seed)
            tw_dist = lda_model.load_word_topics()
            self.lda_tw_dist.append(tw_dist)
            topic_tuples = lda_model.show_topics(num_topics=self.lda_k, num_words=self.nlda_topic_depth, formatted=False)
            lda_topics = [[w for (w, _) in topic[1]] for topic in topic_tuples]
            if self.lda_save_path:
                save_topics(lda_topics, self.lda_save_path + 'topics_{}_{}.csv'.format(self.k, t))
            self.lda_topics.append(lda_topics)
            self.last_lda_beta = lda_model.load_beta()
            self.last_lda_alpha_array_file = lda_model.falphaarrayfile()
            self.last_lda_tw_dist_file = lda_model.fwordweights()

        topic_sets = self.test_all_phi_and_depths(tw_dist, noise_dist, lda_topics)
        return topic_sets, noise_dist

    def __str__(self):
        return 'deNLDA'



from datetime import datetime, timedelta


def load_flat_dataset(path, delimiter=' '):
    dataset = []
    with open(path, 'r') as f:
        for line in f:
            dataset.append(line.strip().split(delimiter))
    return dataset


def save_flat_dataset(path, dataset, delimiter=' '):
    with open(path, 'w') as f:
        for d in dataset:
            f.write('{}\n'.format(delimiter.join(d)))


def load_dated_dataset(path, date_delimiter='\t', doc_delimiter=' '):
    dataset = []
    with open(path, 'r') as f:
        for line in f:
            data = line.strip().split(date_delimiter)
            date, doc = data[0], data[1]
            doc = doc.split(doc_delimiter)
            dataset.append((date, doc))
    return dataset


def save_split_dataset(path, file_name, dataset, delimiter=' '):
    '''

    :param path: PATH SHOULD BE PATH AND NOT A FILE NAME
    :param file_name: the name of the data set
    :param dataset:
    :param delimiter:
    :return:
    '''
    for i in range(0, len(dataset)):
        with open('{}{}_{}.csv'.format(path, file_name, i), 'w') as f:
            for d in dataset[i]:
                f.write('{}\n'.format(delimiter.join(d)))


def load_split_dataset(path, file_name, num_time_periods, delimiter=' '):
    dataset = []
    for i in range(0, num_time_periods):
        docs = []
        with open('{}{}_{}.csv'.format(path, file_name, i), 'r') as f:
            for line in f:
                docs.append(line.strip().split(delimiter))
        dataset.append(docs)
    return dataset


def year(d):
    return datetime(year=d.year, month=1, day=1)


def month(d):
    return datetime(year=d.year, month=d.month, day=1)


def week(d):
    diff = d.weekday()
    return d if diff == 6 else d - timedelta(days=diff + 1)


def fortnight(d):
    return d.isocalendar()[1] - (1 + d.isocalendar()[1] % 2)


def day(d):
    return d.date()


def none(d):
    return 1


def get_time_periods(dataset, epoch_function):
    times = list(set([epoch_function(x[0]) for x in dataset]))
    return sorted(times)


def split_dataset_by_date(dataset, epoch_function):
    split_dataset = []
    time_periods = get_time_periods(dataset, epoch_function)
    [split_dataset.append([]) for x in time_periods]
    for d in dataset:
        split_dataset[time_periods.index(epoch_function[d[0]])].append(d[1])
    return split_dataset


def get_vocabulary(docs):
    '''
    This version of get_vocabulary takes 0.08 seconds on 100,000 documents whereas the old version took forever.
    '''
    vocab = []
    for i in range(0, len(docs)):
        vocab.extend(docs[i])
    return list(set(vocab))


def save_topics(topics, path, delimiter=','):
    with open(path, 'w') as f:
        for topic in topics:
            f.write('{}\n'.format(delimiter.join(topic)))


def save_noise_dist(noise, path):
    with open(path, 'w') as f:
        for word, freq in noise:
            f.write('{},{}\n'.format(word, freq))


def load_topics(path, delimiter=','):
    topics = []
    with open(path, 'r') as f:
        for line in f:
            words = line.strip().split(delimiter)
            words = [w for w in words if len(w) > 0]
            topics.append(words)
    return topics


def load_noise_dist(path, delimiter=','):
    noise_dist = []
    with open(path, 'r') as f:
        for line in f:
            word, freq = line.strip().split(delimiter)
            noise_dist.append((word, freq))
    return noise_dist



#!/home/rob/.env/topics/bin/python
from gensim import corpora
from ..helpers.exceptions import MissingEmbeddingPathError
from ..wrappers import eTNDMallet
from .tnd import TND


class eTND(TND):
    topics = None

    def __init__(self, dataset=None, k=30, alpha=50, beta0=0.01, beta1=25, noise_words_max=200,
                 iterations=1000, top_words=20, topic_word_distribution=None,
                 noise_distribution=None, corpus=None, dictionary=None,
                 save_path=None, mallet_path=None, embedding_path=None,
                 closest_x_words=3, random_seed=1824, run=True, workers=4):

        super().__init__(dataset=dataset, k=k, alpha=alpha, beta0=beta0, beta1=beta1, noise_words_max=noise_words_max,
                         iterations=iterations, top_words=top_words, topic_word_distribution=topic_word_distribution,
                         noise_distribution=noise_distribution, corpus=corpus, dictionary=dictionary,
                         save_path=save_path, mallet_path=mallet_path, random_seed=random_seed, run=False,
                         workers=workers)

        if save_path is not None:
            self.save_path = save_path
        else:
            self.save_path = 'etnd_results/'
        self.embedding_path = embedding_path
        if self.embedding_path is None:
            raise MissingEmbeddingPathError
        self.closest_x_words = closest_x_words

        if run:
            if (self.dataset is not None) and (self.corpus is None or self.dictionary is None):
                self.prepare_data()
            if self.noise_distribution is None:
                self.compute_tnd()

    def compute_tnd(self):
        """
        takes dataset, tnd parameters, tnd mallet path, and computes tnd model on dataset
        sets self.noise_distribution to the noise distribution computed in tnd
        :return: void
        """
        model = eTNDMallet(self.mallet_path, self.corpus, num_topics=self.k, id2word=self.dictionary,
                           workers=self.workers, alpha=self.alpha, beta=self.beta0, skew=self.beta1,
                           iterations=self.iterations, noise_words_max=self.noise_words_max,
                           random_seed=self.random_seed, closest_x_words=self.closest_x_words,
                           embedding_path=self.embedding_path)
        noise = model.load_noise_dist()
        self.noise_distribution = noise
        self.topic_word_distribution = model.load_word_topics()
        topics = model.show_topics(num_topics=self.k, num_words=self.top_words, formatted=False)
        self.topics = [[w for (w, _) in topics[i][1]] for i in range(0, len(topics))]



#!/home/rob/.env/topics/bin/python
import math
import random
from gensim import corpora
from ..helpers.exceptions import MissingModelError, MissingDataSetError
from ..wrappers import TNDMallet


class TND:
    topics = None

    def __init__(self, dataset=None, k=30, alpha=50, beta0=0.01, beta1=25, noise_words_max=200,
                 iterations=1000, top_words=20, topic_word_distribution=None,
                 noise_distribution=None, corpus=None, dictionary=None,
                 save_path=None, mallet_path=None, random_seed=1824, run=True,
                 workers=4):
        self.dataset = dataset
        self.k = k
        self.alpha = alpha
        self.beta0 = beta0
        self.beta1 = beta1
        self.noise_words_max = noise_words_max
        self.iterations = iterations
        self.top_words = top_words
        self.noise_distribution = noise_distribution
        self.topic_word_distribution = topic_word_distribution
        self.corpus = corpus
        self.dictionary = dictionary
        if save_path is not None:
            self.save_path = save_path
        else:
            self.save_path = 'tnd_results/'
        self.mallet_path = mallet_path
        self.random_seed = random_seed
        random.seed(self.random_seed)
        self.workers = workers

        if self.mallet_path is None and self.noise_distribution is None:
            raise MissingModelError('tnd')
        if self.dataset is None and (self.corpus is None or self.dictionary is None):
            raise MissingDataSetError

        if run:
            if (self.dataset is not None) and (self.corpus is None or self.dictionary is None):
                self.prepare_data()
            if self.noise_distribution is None:
                self.compute_tnd()

    def prepare_data(self):
        """
        takes dataset, sets self.dictionary and self.corpus for use in Mallet models and NLDA
        :return: void
        """
        dictionary = corpora.Dictionary(self.dataset)
        dictionary.filter_extremes()
        corpus = [dictionary.doc2bow(doc) for doc in self.dataset]
        self.dictionary = dictionary
        self.corpus = corpus

    def compute_tnd(self):
        """
        takes dataset, tnd parameters, tnd mallet path, and computes tnd model on dataset
        sets self.noise_distribution to the noise distribution computed in tnd
        :return: void
        """
        model = TNDMallet(self.mallet_path, self.corpus, num_topics=self.k, id2word=self.dictionary,
                          workers=self.workers,
                          alpha=self.alpha, beta=self.beta0, skew=self.beta1,
                          iterations=self.iterations, noise_words_max=self.noise_words_max,
                          random_seed=self.random_seed)
        noise = model.load_noise_dist()
        self.noise_distribution = noise
        self.topic_word_distribution = model.load_word_topics()
        topics = model.show_topics(num_topics=self.k, num_words=self.top_words, formatted=False)
        self.topics = [[w for (w, _) in topics[i][1]] for i in range(0, len(topics))]

    def get_topics(self, top_words=None):
        """
        takes top_words and self.topics, returns a list of topic lists of length top_words
        :param top_words:
        :return: list of topic lists
        """
        if top_words is None:
            top_words = self.top_words
        topics = self.topics
        if topics is None or len(topics) < 1:
            raise ValueError('No topics have been computed yet.')

        return [x[:top_words] for x in topics]

    def get_noise_distribution(self, tnd_noise_words_max=None):
        """
        takes self.noise_distribution and tnd_noise_words_max
        returns a list of (noise word, frequency) tuples ranked by frequency
        :param tnd_noise_words_max:
        :return: list of (noise word, frequency) tuples
        """
        if tnd_noise_words_max is None:
            tnd_noise_words_max = self.noise_words_max
        noise = self.noise_distribution
        if noise is None or len(noise) < 1:
            raise ValueError('No noise distribution has been computed yet.')

        noise_list = sorted([(x, int(noise[x])) for x in noise.keys()], key=lambda x: x[1], reverse=True)
        return noise_list[:tnd_noise_words_max]



#!/home/rob/.env/topics/bin/python
import math
import random
from gensim import corpora
from ..helpers.exceptions import MissingDataSetError
from ..wrappers import dLDAMallet, dTNDMallet
from ..helpers.common import save_topics, save_noise_dist
from .nlda import NLDA


class dNLDA(NLDA):
    topics = []

    def __init__(self, dataset=None, tnd_k=30, tnd_alpha=50, tnd_beta0=0.01, tnd_beta1=25, tnd_noise_words_max=200,
                 tnd_iterations=1000, lda_iterations=1000, lda_k=30, lda_beta=0.01, nlda_phi=10,
                 nlda_topic_depth=100, top_words=20, nlda_phi_list=None, nlda_topic_depth_list=None,
                 num_time_periods=10, tnd_noise_distribution=[], lda_tw_dist=[], lda_topics=[], corpus=None,
                 dictionary=None, last_tnd_alpha_array_file=None, last_tnd_noise_dist_file=None,
                 last_tnd_tw_dist_file=None, last_lda_alpha_array_file=None, last_lda_tw_dist_file=None,
                 lda_save_path=None, save_path=None, mallet_tnd_path=None, mallet_lda_path=None, random_seed=1824,
                 tnd_workers=4, lda_workers=4, run=True):

        super().__init__(dataset=dataset, tnd_k=tnd_k, tnd_alpha=tnd_alpha, tnd_beta0=tnd_beta0, tnd_beta1=tnd_beta1,
                         tnd_noise_words_max=tnd_noise_words_max, tnd_iterations=tnd_iterations,
                         lda_iterations=lda_iterations, lda_k=lda_k, nlda_phi=nlda_phi,
                         nlda_topic_depth=nlda_topic_depth, top_words=top_words,
                         tnd_noise_distribution=tnd_noise_distribution, lda_tw_dist=lda_tw_dist, lda_topics=lda_topics,
                         corpus=corpus, dictionary=dictionary, save_path=save_path, mallet_tnd_path=mallet_tnd_path,
                         mallet_lda_path=mallet_lda_path, random_seed=random_seed, run=False, tnd_workers=tnd_workers,
                         lda_workers=lda_workers)
        if save_path is not None:
            save = True
            self.save_path = save_path
        else:
            save = False
            self.save_path = 'dnlda_results/'
        self.last_lda_beta = lda_beta
        self.last_lda_alpha_array_file = last_lda_alpha_array_file
        self.last_lda_tw_dist_file = last_lda_tw_dist_file
        self.last_tnd_beta = tnd_beta0
        self.last_tnd_alpha_array_file = last_tnd_alpha_array_file
        self.last_tnd_tw_dist_file = last_tnd_tw_dist_file
        self.last_tnd_noise_dist_file = last_tnd_noise_dist_file
        self.lda_save_path = lda_save_path
        self.nlda_phi_list = nlda_phi_list
        self.nlda_topic_depth_list = nlda_topic_depth_list

        self.num_time_periods = num_time_periods
        if self.dataset is not None:
            self.num_time_periods = len(dataset)
        if self.dataset is None and (self.corpus is None or self.dictionary is None
                                     or len(self.tnd_noise_distribution) != self.num_time_periods or
                                     len(self.lda_tw_dist) != self.num_time_periods):
            raise MissingDataSetError

        if run:
            self.run_nlda_all_time_periods(save=save)

    def prepare_data(self, t):
        """
        takes dataset, sets self.dictionary and self.corpus for use in Mallet models and NLDA
        :return: void
        """
        dictionary = corpora.Dictionary(self.dataset[t])
        dictionary.filter_extremes()
        corpus = [dictionary.doc2bow(doc) for doc in self.dataset[t]]
        self.dictionary = dictionary
        self.corpus = corpus

    def compute_nlda(self, noise_dist, tw_dist, lda_topics, phi, depth):
        """
        takes self.tnd_noise_distribution, self.lda_tw_dist, self.phi, self.top_words, and computes NLDA topics
        sets self.topics to the set of topics computed from noise distribution and topic word distribution
        :return: void
        """
        topics = []
        for i in range(0, len(lda_topics)):
            topic = lda_topics[i]
            final_topic = []
            j = 0
            while len(topic) > j and len(final_topic) < self.top_words and j < depth:
                w = topic[j]
                token_id = self.dictionary.token2id[w]
                beta = 2
                if w in noise_dist:
                    beta += noise_dist[w]
                beta = max(2, beta * (phi / self.k))
                alpha = 2 + tw_dist[i, token_id]
                roll = random.betavariate(alpha=math.sqrt(alpha), beta=math.sqrt(beta))
                if roll >= 0.5:
                    final_topic.append(w)
                    if w not in noise_dist:
                        noise_dist[w] = 0
                    noise_dist[w] += (alpha - 2)
                j += 1
            topics.append(final_topic)
        return topics

    def test_all_phi_and_depths(self, tw_dist, noise_dist, lda_topics):
        phi = self.nlda_phi
        depth = self.nlda_topic_depth
        phi_depth_combos = []
        if self.nlda_phi_list is not None and self.nlda_topic_depth_list is not None:
            for i in range(0, len(self.nlda_phi_list)):
                phi = self.nlda_phi_list[i]
                for j in range(0, len(self.nlda_topic_depth_list)):
                    depth = self.nlda_topic_depth_list[j]
                    phi_depth_combos.append((phi, depth))
        elif self.nlda_phi_list is None and self.nlda_topic_depth_list is not None:
            for j in range(0, len(self.nlda_topic_depth_list)):
                depth = self.nlda_topic_depth_list[j]
                phi_depth_combos.append((phi, depth))
        elif self.nlda_phi_list is not None and self.nlda_topic_depth_list is None:
            for i in range(0, len(self.nlda_phi_list)):
                phi = self.nlda_phi_list[i]
                phi_depth_combos.append((phi, depth))
        else:
            phi_depth_combos.append((phi, depth))

        topic_sets = []
        for phi, depth in phi_depth_combos:
            topics = self.compute_nlda(noise_dist=noise_dist, tw_dist=tw_dist, lda_topics=lda_topics, phi=phi,
                                       depth=depth)
            topic_sets.append((phi, depth, topics))
        return topic_sets

    def run_nlda_one_time_period(self, t):
        self.prepare_data(t)
        if len(self.tnd_noise_distribution) > t:
            noise_dist = self.tnd_noise_distribution[t]
        else:
            tnd_model = dTNDMallet(self.mallet_tnd_path, corpus=self.corpus, num_topics=self.tnd_k,
                                   beta=self.last_tnd_beta, id2word=self.dictionary, iterations=self.tnd_iterations,
                                   skew=self.tnd_beta1, noise_words_max=self.tnd_noise_words_max,
                                   workers=self.tnd_workers, noise_dist_file=self.last_tnd_noise_dist_file,
                                   tw_dist_file=self.last_tnd_tw_dist_file,
                                   alpha_array_infile=self.last_tnd_alpha_array_file, alpha=self.tnd_alpha)
            noise_dist = tnd_model.load_noise_dist()
            self.tnd_noise_distribution.append(noise_dist)
            self.last_tnd_beta = tnd_model.load_beta()
            self.last_tnd_alpha_array_file = tnd_model.falphaarrayfile()
            self.last_tnd_noise_dist_file = tnd_model.fnoisefile()
            self.last_tnd_tw_dist_file = tnd_model.fwordweights()

        if len(self.lda_tw_dist) > t and len(self.lda_topics) > t:
            tw_dist = self.lda_tw_dist[t]
            lda_topics = self.lda_topics[t]
        else:
            lda_model = dLDAMallet(self.mallet_lda_path, self.corpus, num_topics=self.lda_k, id2word=self.dictionary,
                               workers=4, tw_dist_file=self.last_lda_tw_dist_file,
                               alpha_array_infile=self.last_lda_alpha_array_file, beta=self.last_lda_beta,
                               iterations=self.lda_iterations, random_seed=self.random_seed)
            tw_dist = lda_model.load_word_topics()
            self.lda_tw_dist.append(tw_dist)
            topic_tuples = lda_model.show_topics(num_topics=self.lda_k, num_words=self.nlda_topic_depth,
                                                 formatted=False)
            lda_topics = [[w for (w, _) in topic[1]] for topic in topic_tuples]
            if self.lda_save_path:
                save_topics(lda_topics, self.lda_save_path + 'topics_{}_{}.csv'.format(self.lda_k, t))
            self.lda_topics.append(lda_topics)
            self.last_lda_beta = lda_model.load_beta()
            self.last_lda_alpha_array_file = lda_model.falphaarrayfile()
            self.last_lda_tw_dist_file = lda_model.fwordweights()

        topic_sets = self.test_all_phi_and_depths(tw_dist, noise_dist, lda_topics)
        return topic_sets, noise_dist


    def run_nlda_all_time_periods(self, save=True):
        for t in range(0, len(self.dataset)):
            topic_sets, noise_dist = self.run_nlda_one_time_period(t)
            self.topics.append(topic_sets[0])
            if save:
                for phi, depth, topics in topic_sets:
                    save_topics(topics, self.save_path + 'topics_{}_{}_{}_{}.csv'.format(self.k, t, phi, depth))
                noise_list = sorted([(x, self.tnd_noise_distribution[t][x]) for x in self.tnd_noise_distribution[t].keys()],
                                    key=lambda x: x[1], reverse=True)
                save_noise_dist(noise_list, self.save_path + 'noise_{}_{}.csv'.format(self.k, t))


    def get_topics(self, t, top_words=None):
        """
        takes top_words and self.topics, returns a list of topic lists of length top_words
        :param t: time period
        :param top_words:
        :return: list of topic lists
        """
        if top_words is None:
            top_words = self.top_words
        topics = self.topics
        if len(topics) < 1:
            raise ValueError('No topics have been computed yet.')
        elif len(topics[t]) < 1:
            raise ValueError('No topics have been computed for this time period yet.')

        return [x[:top_words] for x in topics[t]]

    def get_noise_distribution(self, t, tnd_noise_words_max=None):
        """
        takes self.tnd_noise_distribution and tnd_noise_words_max
        returns a list of (noise word, frequency) tuples ranked by frequency
        :param t: time period
        :param tnd_noise_words_max:
        :return: list of (noise word, frequency) tuples
        """
        if tnd_noise_words_max is None:
            tnd_noise_words_max = self.tnd_noise_words_max
        if len(self.tnd_noise_distribution) < 1:
            raise ValueError('No noise distribution has been computed yet.')
        elif len(self.tnd_noise_distribution[t]) < 1:
            raise ValueError('No noise distribution has been computed for this time period yet.')

        noise = self.tnd_noise_distribution[t]
        noise_list = sorted([(x, int(noise[x])) for x in noise.keys()], key=lambda x: x[1], reverse=True)
        return noise_list[:tnd_noise_words_max]

    def __str__(self):
        return 'dNLDA'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 Radim Rehurek <radimrehurek@seznam.cz>
# Licensed under the GNU LGPL v2.1 - http://www.gnu.org/licenses/lgpl.html


r"""Python wrapper for `Dynamic Latent Dirichlet Allocation (dLDA) <https://en.wikipedia.org/wiki/Latent_Dirichlet_allocation>`_
from `MALLET, the Java topic modelling toolkit <http://mallet.cs.umass.edu/>`_

This module allows both LDA model estimation from a training corpus and inference of topic distribution on new,
unseen documents, using an (optimized version of) collapsed gibbs sampling from MALLET.

Notes
-----
MALLET's LDA training requires :math:`O(corpus\_words)` of memory, keeping the entire corpus in RAM.
If you find yourself running out of memory, either decrease the `workers` constructor parameter,
or use :class:`gensim.models.ldamodel.LdaModel` or :class:`gensim.models.ldamulticore.LdaMulticore`
which needs only :math:`O(1)` memory.
The wrapped model can NOT be updated with new documents for online training -- use
:class:`~gensim.models.ldamodel.LdaModel` or :class:`~gensim.models.ldamulticore.LdaMulticore` for that.

Installation
------------
Use `official guide <http://mallet.cs.umass.edu/download.php>`_ or this one ::

    sudo apt-get install default-jdk
    sudo apt-get install ant
    git clone git@github.com:mimno/Mallet.git
    cd Mallet/
    ant

Examples
--------
.. sourcecode:: pycon

    >>> from gensim.test.utils import common_corpus, common_dictionary
    >>> from gensim.models.wrappers import LdaMallet
    >>>
    >>> path_to_mallet_binary = "/path/to/mallet/binary"
    >>> model = LdaMallet(path_to_mallet_binary, corpus=common_corpus, num_topics=20, id2word=common_dictionary)
    >>> vector = model[common_corpus[0]]  # LDA topics of a documents

"""


import logging
import os
import random
import tempfile

from gensim import utils
from gensim.utils import check_output

from .base_wrapper import BaseMalletWrapper

logger = logging.getLogger(__name__)


class dLDAMallet(BaseMalletWrapper):
    """Python wrapper for LDA using `MALLET <http://mallet.cs.umass.edu/>`_.

    Communication between MALLET and Python takes place by passing around data files on disk
    and calling Java with subprocess.call().

    Warnings
    --------
    This is **only** python wrapper for `MALLET LDA <http://mallet.cs.umass.edu/>`_,
    you need to install original implementation first and pass the path to binary to ``mallet_path``.

    """
    def __init__(self, mallet_path, corpus=None, num_topics=100, alpha=50, beta=0.01, id2word=None, workers=4, prefix=None,
                 optimize_interval=0, iterations=1000, topic_threshold=0.0, random_seed=0,
                 alpha_array_infile=None, tw_dist_file=None):
        """

        Parameters
        ----------
        mallet_path : str
            Path to the mallet binary, e.g. `/home/username/mallet-2.0.7/bin/mallet`.
        corpus : iterable of iterable of (int, int), optional
            Collection of texts in BoW format.
        num_topics : int, optional
            Number of topics.
        alpha : int, optional
            Alpha parameter of LDA.
        id2word : :class:`~gensim.corpora.dictionary.Dictionary`, optional
            Mapping between tokens ids and words from corpus, if not specified - will be inferred from `corpus`.
        workers : int, optional
            Number of threads that will be used for training.
        prefix : str, optional
            Prefix for produced temporary files.
        optimize_interval : int, optional
            Optimize hyperparameters every `optimize_interval` iterations
            (sometimes leads to Java exception 0 to switch off hyperparameter optimization).
        iterations : int, optional
            Number of training iterations.
        topic_threshold : float, optional
            Threshold of the probability above which we consider a topic.
        random_seed: int, optional
            Random seed to ensure consistent results, if 0 - use system clock.

        """
        super().__init__(mallet_path, corpus=corpus, num_topics=num_topics, alpha=alpha, beta=beta, id2word=id2word,
                         workers=workers, prefix=prefix, optimize_interval=optimize_interval, iterations=iterations,
                         topic_threshold=topic_threshold, random_seed=random_seed)
        self.alpha_array_infile = alpha_array_infile
        self.tw_dist_file = tw_dist_file
        if corpus is not None:
            self.train(corpus)


    def falphaarrayfile(self):
        '''
        Get path to noise file

        :return:
        ---------
        str
        |   Path to noise distribution file.
        '''
        return self.prefix + 'alpha_array.csv'

    def fbetafile(self):
        '''
        Get path to noise file

        :return:
        ---------
        str
        |   Path to noise distribution file.
        '''
        return self.prefix + 'beta.csv'

    def train(self, corpus):
        """Train Mallet LDA.

        Parameters
        ----------
        corpus : iterable of iterable of (int, int)
            Corpus in BoW format

        """
        self.convert_input(corpus, infer=False)
        cmd = self.mallet_path + ' train-topics --input %s --num-topics %s  --alpha %s --optimize-interval %s '\
            '--num-threads %s --output-state %s --output-doc-topics %s --output-topic-keys %s '\
            '--num-iterations %s --inferencer-filename %s --doc-topics-threshold %s  --random-seed %s '\
            '--topic-word-weights-file %s --output-alpha-array %s --beta %s --output-beta %s '

        cmd = cmd % (
            self.fcorpusmallet(), self.num_topics, self.alpha, self.optimize_interval,
            self.workers, self.fstate(), self.fdoctopics(), self.ftopickeys(), self.iterations,
            self.finferencer(), self.topic_threshold, str(self.random_seed), self.fwordweights(),
            self.falphaarrayfile(), str(self.beta), self.fbetafile()
        )
        if self.alpha_array_infile is not None:
            cmd += '--alpha-array-infile %s ' % self.alpha_array_infile
        else:
            cmd += '--alpha %s ' % str(self.alpha)
        if self.tw_dist_file is not None:
            cmd += '--tw-dist-infile %s ' % self.tw_dist_file
        # NOTE "--keep-sequence-bigrams" / "--use-ngrams true" poorer results + runs out of memory
        logger.info("training MALLET LDA with %s", cmd)
        check_output(args=cmd, shell=True)
        self.word_topics = self.load_word_topics()

    def load_alpha_array(self):
        alpha = []
        logger.info("loading alpha array from %s", self.falphaarrayfile())
        with open(self.falphaarrayfile(), 'r') as f:
            for line in f:
                prob = line.strip()
                alpha.append(float(prob))
        return alpha

    def load_beta(self):
        beta = 0
        logger.info("loading beta from %s", self.fbetafile())
        with open(self.fbetafile(), 'r') as f:
            for line in f:
                prob = line.strip()
                beta = float(prob)
        return beta



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 Radim Rehurek <radimrehurek@seznam.cz>
# Licensed under the GNU LGPL v2.1 - http://www.gnu.org/licenses/lgpl.html
# Edited to run Guided Topic Model and save noise distribution by Rob Churchill (March 2021)

r"""Python wrapper for `Latent Dirichlet Allocation (LDA) <https://en.wikipedia.org/wiki/Latent_Dirichlet_allocation>`_
from `MALLET, the Java topic modelling toolkit <http://mallet.cs.umass.edu/>`_

This module allows both LDA model estimation from a training corpus and inference of topic distribution on new,
unseen documents, using an (optimized version of) collapsed gibbs sampling from MALLET.

Notes
-----
MALLET's LDA training requires :math:`O(corpus\_words)` of memory, keeping the entire corpus in RAM.
If you find yourself running out of memory, either decrease the `workers` constructor parameter,
or use :class:`gensim.models.ldamodel.LdaModel` or :class:`gensim.models.ldamulticore.LdaMulticore`
which needs only :math:`O(1)` memory.

"""


import logging
from gensim.utils import check_output

from ..helpers.exceptions import MissingSeedWeightsError
from .base_wrapper import BaseMalletWrapper

logger = logging.getLogger(__name__)


class GTMMallet(BaseMalletWrapper):
    """Python wrapper for Seeded LDA using `MALLET <http://mallet.cs.umass.edu/>`_.

    Communication between MALLET and Python takes place by passing around data files on disk
    and calling Java with subprocess.call().

    Warnings
    --------
    This is **only** python wrapper for `MALLET LDA <http://mallet.cs.umass.edu/>`_,
    you need to install original implementation first and pass the path to binary to ``mallet_path``.

    """
    def __init__(self, mallet_path, corpus=None, num_topics=100, alpha=50, beta=0.01, id2word=None, workers=4,
                 prefix=None, optimize_interval=0, iterations=1000, topic_threshold=0.0, random_seed=0,
                 seed_topics_file=None, sampling_scheme=2, over_sampling_factor=1, seed_gpu_weights=None):
        """

        Parameters
        ----------
        mallet_path : str
            Path to the mallet binary, e.g. `/home/username/mallet-2.0.7/bin/mallet`.
        corpus : iterable of iterable of (int, int), optional
            Collection of texts in BoW format.
        num_topics : int, optional
            Number of topics.
        alpha : int, optional
            Alpha parameter of LDA.
        id2word : :class:`~gensim.corpora.dictionary.Dictionary`, optional
            Mapping between tokens ids and words from corpus, if not specified - will be inferred from `corpus`.
        workers : int, optional
            Number of threads that will be used for training.
        prefix : str, optional
            Prefix for produced temporary files.
        optimize_interval : int, optional
            Optimize hyperparameters every `optimize_interval` iterations
            (sometimes leads to Java exception 0 to switch off hyperparameter optimization).
        iterations : int, optional
            Number of training iterations.
        topic_threshold : float, optional
            Threshold of the probability above which we consider a topic.
        random_seed: int, optional
            Random seed to ensure consistent results, if 0 - use system clock.
        sampling_scheme: int, optional
            0: normal sampling
            1: oversampling of observed seed word
            2: GPU sampling of seed words
        over_sampling_factor: double, optional
            the amount to oversample a seed word by in sampling scheme 1.
            A value of 10 will oversample a seed word 10:1
        seed_gpu_weights: list of lists corresponding to one weight per seed word per topic

        """
        super().__init__(mallet_path, corpus=corpus, num_topics=num_topics, alpha=alpha, beta=beta, id2word=id2word,
                         workers=workers, prefix=prefix, optimize_interval=optimize_interval, iterations=iterations,
                         topic_threshold=topic_threshold, random_seed=random_seed)
        self.seed_topics_file = seed_topics_file
        self.sampling_scheme = sampling_scheme
        self.over_sampling_factor = over_sampling_factor
        self.seed_gpu_weights = seed_gpu_weights
        if self.seed_gpu_weights is None:
            raise MissingSeedWeightsError
        self.save_seed_gpu_weights()
        if corpus is not None:
            self.train(corpus)

    def fnoisefile(self):
        '''
        Get path to noise file

        :return:
        ---------
        str
        |   Path to noise distribution file.
        '''
        return self.prefix + 'noise_dist.csv'

    def fseedgpuweights(self):
        """Get path to word weight file.

        Returns
        -------
        str
            Path to word weight file.

        """
        return self.prefix + 'seedgpuweights.txt'

    def save_seed_gpu_weights(self):
        with open(self.fseedgpuweights(), 'w') as f:
            if self.seed_gpu_weights is not None:
                for weights in self.seed_gpu_weights:
                    w_strs = ["%.2f" % x for x in weights]
                    f.write('{}\n'.format(','.join(w_strs)))

    def train(self, corpus):
        """Train Mallet LDA.

        Parameters
        ----------
        corpus : iterable of iterable of (int, int)
            Corpus in BoW format

        """
        self.convert_input(corpus, infer=False)
        cmd = self.mallet_path + ' train-topics --input %s --num-topics %s  --alpha %s --optimize-interval %s '\
            '--num-threads %s --output-state %s --output-doc-topics %s --output-topic-keys %s '\
            '--num-iterations %s --inferencer-filename %s --doc-topics-threshold %s  --random-seed %s --beta %s ' \
            '--samplingScheme %s --overSamplingFactor %s --seed-gpu-weights %s'

        cmd = cmd % (
            self.fcorpusmallet(), self.num_topics, self.alpha, self.optimize_interval,
            self.workers, self.fstate(), self.fdoctopics(), self.ftopickeys(), self.iterations,
            self.finferencer(), self.topic_threshold, str(self.random_seed), str(self.beta),
            str(self.sampling_scheme), str(self.over_sampling_factor), self.fseedgpuweights()
        )
        if self.seed_topics_file is not None:
            cmd += '  --seed-topics %s' % self.seed_topics_file
        # NOTE "--keep-sequence-bigrams" / "--use-ngrams true" poorer results + runs out of memory
        logger.info("training MALLET LDA with %s", cmd)
        check_output(args=cmd, shell=True)
        self.word_topics = self.load_word_topics()
        # NOTE - we are still keeping the wordtopics variable to not break backward compatibility.
        # word_topics has replaced wordtopics throughout the code;
        # wordtopics just stores the values of word_topics when train is called.
        self.wordtopics = self.word_topics



#!/home/rob/.env/topics/bin/python
import math
import random
from gensim import corpora
from ..helpers.exceptions import MissingModelError, MissingDataSetError, MissingEmbeddingPathError
from ..wrappers import LDAMallet, eTNDMallet
from .nlda import NLDA


class eNLDA(NLDA):
    topics = None

    def __init__(self, dataset=None, tnd_k=30, tnd_alpha=50, tnd_beta0=0.01, tnd_beta1=25, tnd_noise_words_max=200,
                 tnd_iterations=1000, lda_iterations=1000, lda_k=30, nlda_phi=10, nlda_topic_depth=100, top_words=20,
                 tnd_noise_distribution=None, lda_tw_dist=None, lda_topics=None, corpus=None, dictionary=None,
                 save_path=None, mallet_tnd_path=None, mallet_lda_path=None, embedding_path=None,
                 closest_x_words=3, random_seed=1824, run=True, tnd_workers=4, lda_workers=4):

        super().__init__(dataset=dataset, tnd_k=tnd_k, tnd_alpha=tnd_alpha, tnd_beta0=tnd_beta0, tnd_beta1=tnd_beta1,
                         tnd_noise_words_max=tnd_noise_words_max, tnd_iterations=tnd_iterations,
                         lda_iterations=lda_iterations, lda_k=lda_k, nlda_phi=nlda_phi,
                         nlda_topic_depth=nlda_topic_depth, top_words=top_words,
                         tnd_noise_distribution=tnd_noise_distribution, lda_tw_dist=lda_tw_dist, lda_topics=lda_topics,
                         corpus=corpus, dictionary=dictionary, save_path=save_path, mallet_tnd_path=mallet_tnd_path,
                         mallet_lda_path=mallet_lda_path, random_seed=random_seed, run=False, tnd_workers=tnd_workers,
                         lda_workers=lda_workers)

        if save_path is not None:
            self.save_path = save_path
        else:
            self.save_path = 'nlda_results/'
        self.mallet_tnd_path = mallet_tnd_path
        self.mallet_lda_path = mallet_lda_path
        self.embedding_path = embedding_path
        if self.embedding_path is None:
            raise MissingEmbeddingPathError
        self.closest_x_words = closest_x_words

        if run:
            if (self.dataset is not None) and (self.corpus is None or self.dictionary is None):
                self.prepare_data()
            if self.tnd_noise_distribution is None:
                self.compute_tnd()
            if self.lda_tw_dist is None:
                self.compute_lda()

            self.compute_nlda()

    def compute_tnd(self):
        """
        takes dataset, tnd parameters, tnd mallet path, and computes tnd model on dataset
        sets self.tnd_noise_distribution to the noise distribution computed in tnd
        :return: void
        """
        model = eTNDMallet(self.mallet_tnd_path, self.corpus, num_topics=self.tnd_k, id2word=self.dictionary,
                           workers=self.tnd_workers, alpha=self.tnd_alpha, beta=self.tnd_beta0, skew=self.tnd_beta1,
                           iterations=self.tnd_iterations, noise_words_max=self.tnd_noise_words_max,
                           random_seed=self.random_seed, closest_x_words=self.closest_x_words,
                           embedding_path=self.embedding_path)
        noise = model.load_noise_dist()
        self.tnd_noise_distribution = noise





def compute_idf_weights(dataset, topics):
    weights = []
    num = len(dataset)
    for topic in topics:
        weight_list = []
        for w in topic:
            df = 1 + sum([1 for d in dataset if w in d])
            weight_list.append(num/df)
        weights.append(weight_list)
    return weights



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 Radim Rehurek <radimrehurek@seznam.cz>
# Licensed under the GNU LGPL v2.1 - http://www.gnu.org/licenses/lgpl.html
# Edited to run dynamic TND and save noise distribution by Rob Churchill (Feb 2021)

r"""Python wrapper for `Latent Dirichlet Allocation (LDA) <https://en.wikipedia.org/wiki/Latent_Dirichlet_allocation>`_
from `MALLET, the Java topic modelling toolkit <http://mallet.cs.umass.edu/>`_

This module allows both LDA model estimation from a training corpus and inference of topic distribution on new,
unseen documents, using an (optimized version of) collapsed gibbs sampling from MALLET.

Notes
-----
MALLET's LDA training requires :math:`O(corpus\_words)` of memory, keeping the entire corpus in RAM.
If you find yourself running out of memory, either decrease the `workers` constructor parameter,
or use :class:`gensim.models.ldamodel.LdaModel` or :class:`gensim.models.ldamulticore.LdaMulticore`
which needs only :math:`O(1)` memory.

"""


import logging
import os
import random
import warnings
import tempfile
import xml.etree.ElementTree as et
import zipfile
from itertools import chain

import numpy

from gensim import utils, matutils
from gensim.models import basemodel
from gensim.models.ldamodel import LdaModel
from gensim.utils import check_output, revdict

from .tnd import TNDMallet, eTNDMallet

logger = logging.getLogger(__name__)


class dTNDMallet(TNDMallet):
    """Python wrapper for dTND using `MALLET <http://mallet.cs.umass.edu/>`_.

    Communication between MALLET and Python takes place by passing around data files on disk
    and calling Java with subprocess.call().

    Warnings
    --------
    This is **only** python wrapper for `MALLET LDA <http://mallet.cs.umass.edu/>`_,
    you need to install original implementation first and pass the path to binary to ``mallet_path``.

    """
    def __init__(self, mallet_path, corpus=None, num_topics=100, alpha=50, beta=0.01, id2word=None, workers=4, prefix=None,
                 optimize_interval=0, iterations=1000, topic_threshold=0.0, random_seed=0, noise_words_max=100, skew=0,
                 alpha_array_infile=None, tw_dist_file=None, noise_dist_file=None):
        """

        Parameters
        ----------
        mallet_path : str
            Path to the mallet binary, e.g. `/home/username/mallet-2.0.7/bin/mallet`.
        corpus : iterable of iterable of (int, int), optional
            Collection of texts in BoW format.
        num_topics : int, optional
            Number of topics.
        alpha : int, optional
            Alpha parameter of LDA.
        id2word : :class:`~gensim.corpora.dictionary.Dictionary`, optional
            Mapping between tokens ids and words from corpus, if not specified - will be inferred from `corpus`.
        workers : int, optional
            Number of threads that will be used for training.
        prefix : str, optional
            Prefix for produced temporary files.
        optimize_interval : int, optional
            Optimize hyperparameters every `optimize_interval` iterations
            (sometimes leads to Java exception 0 to switch off hyperparameter optimization).
        iterations : int, optional
            Number of training iterations.
        topic_threshold : float, optional
            Threshold of the probability above which we consider a topic.
        random_seed: int, optional
            Random seed to ensure consistent results, if 0 - use system clock.

        """
        super().__init__(mallet_path, corpus=corpus, num_topics=num_topics, alpha=alpha, beta=beta, id2word=id2word,
                         workers=workers, prefix=prefix, optimize_interval=optimize_interval, iterations=iterations,
                         topic_threshold=topic_threshold, random_seed=random_seed, noise_words_max=noise_words_max,
                         skew=skew, is_parent=True)
        self.alpha_array_infile = alpha_array_infile
        self.tw_dist_file = tw_dist_file
        self.noise_dist_file = noise_dist_file
        if corpus is not None:
            self.train(corpus)

    def ftwdistfile(self):
        '''
        Get path to noise file

        :return:
        ---------
        str
        |   Path to noise distribution file.
        '''
        return self.prefix + 'tw_dist.csv'

    def falphaarrayfile(self):
        '''
        Get path to noise file

        :return:
        ---------
        str
        |   Path to noise distribution file.
        '''
        return self.prefix + 'alpha_array.csv'

    def fbetafile(self):
        '''
        Get path to noise file

        :return:
        ---------
        str
        |   Path to noise distribution file.
        '''
        return self.prefix + 'beta.csv'

    def train(self, corpus):
        """Train Mallet LDA.

        Parameters
        ----------
        corpus : iterable of iterable of (int, int)
            Corpus in BoW format

        """
        self.convert_input(corpus, infer=False)
        cmd = self.mallet_path + ' train-topics --input %s --num-topics %s --optimize-interval %s '\
            '--num-threads %s --output-state %s --output-doc-topics %s --output-topic-keys %s '\
            '--num-iterations %s --inferencer-filename %s --doc-topics-threshold %s  --random-seed %s --output-noise %s '\
            '--noise-words-max %s --skew %s --beta %s --topic-word-weights-file %s --output-alpha-array %s --output-beta %s '

        cmd = cmd % (
            self.fcorpusmallet(), self.num_topics, self.optimize_interval,
            self.workers, self.fstate(), self.fdoctopics(), self.ftopickeys(), self.iterations,
            self.finferencer(), self.topic_threshold, str(self.random_seed), self.fnoisefile(), str(self.noise_words_max),
            str(self.skew), str(self.beta), self.fwordweights(), self.falphaarrayfile(), self.fbetafile()
        )
        if self.alpha_array_infile is not None:
            cmd += '--alpha-array-infile %s ' % self.alpha_array_infile
        else:
            cmd += '--alpha %s ' % str(self.alpha)
        if self.tw_dist_file is not None:
            cmd += '--tw-dist-infile %s ' % self.tw_dist_file
        if self.noise_dist_file is not None:
            cmd += '--noise-dist-infile %s ' % self.noise_dist_file
        # NOTE "--keep-sequence-bigrams" / "--use-ngrams true" poorer results + runs out of memory
        logger.info("training MALLET LDA with %s", cmd)
        check_output(args=cmd, shell=True)
        self.word_topics = self.load_word_topics()
        # NOTE - we are still keeping the wordtopics variable to not break backward compatibility.
        # word_topics has replaced wordtopics throughout the code;
        # wordtopics just stores the values of word_topics when train is called.
        self.wordtopics = self.word_topics

    def load_alpha_array(self):
        alpha = []
        logger.info("loading alpha array from %s", self.falphaarrayfile())
        with open(self.falphaarrayfile(), 'r') as f:
            for line in f:
                prob = line.strip()
                alpha.append(float(prob))
        return alpha

    def load_beta(self):
        beta = 0
        logger.info("loading beta from %s", self.fbetafile())
        with open(self.fbetafile(), 'r') as f:
            for line in f:
                prob = line.strip()
                beta = float(prob)
        return beta

#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 Radim Rehurek <radimrehurek@seznam.cz>
# Licensed under the GNU LGPL v2.1 - http://www.gnu.org/licenses/lgpl.html
# Edited to run dynamic embedded TND and save noise distribution by Rob Churchill (Feb 2021)

r"""Python wrapper for `Latent Dirichlet Allocation (LDA) <https://en.wikipedia.org/wiki/Latent_Dirichlet_allocation>`_
from `MALLET, the Java topic modelling toolkit <http://mallet.cs.umass.edu/>`_

This module allows both LDA model estimation from a training corpus and inference of topic distribution on new,
unseen documents, using an (optimized version of) collapsed gibbs sampling from MALLET.

Notes
-----
MALLET's LDA training requires :math:`O(corpus\_words)` of memory, keeping the entire corpus in RAM.
If you find yourself running out of memory, either decrease the `workers` constructor parameter,
or use :class:`gensim.models.ldamodel.LdaModel` or :class:`gensim.models.ldamulticore.LdaMulticore`
which needs only :math:`O(1)` memory.
The wrapped model can NOT be updated with new documents for online training 

"""

class deTNDMallet(eTNDMallet):
    """Python wrapper for NFT using `MALLET <http://mallet.cs.umass.edu/>`_.

    Communication between MALLET and Python takes place by passing around data files on disk
    and calling Java with subprocess.call().

    Warnings
    --------
    This is **only** python wrapper for `MALLET LDA <http://mallet.cs.umass.edu/>`_,
    you need to install original implementation first and pass the path to binary to ``mallet_path``.

    """
    def __init__(self, mallet_path, corpus=None, num_topics=100, alpha=50, beta=0.01, id2word=None, workers=4, prefix=None,
                 optimize_interval=0, iterations=1000, topic_threshold=0.0, random_seed=0, noise_words_max=100, skew=0,
                 tau=200, embedding_path=None, closest_x_words=3, save_embedding_comp=True,
                 alpha_array_infile=None, tw_dist_file=None, noise_dist_file=None):
        """

        Parameters
        ----------
        mallet_path : str
            Path to the mallet binary, e.g. `/home/username/mallet-2.0.7/bin/mallet`.
        corpus : iterable of iterable of (int, int), optional
            Collection of texts in BoW format.
        num_topics : int, optional
            Number of topics.
        alpha : int, optional
            Alpha parameter of LDA.
        id2word : :class:`~gensim.corpora.dictionary.Dictionary`, optional
            Mapping between tokens ids and words from corpus, if not specified - will be inferred from `corpus`.
        workers : int, optional
            Number of threads that will be used for training.
        prefix : str, optional
            Prefix for produced temporary files.
        optimize_interval : int, optional
            Optimize hyperparameters every `optimize_interval` iterations
            (sometimes leads to Java exception 0 to switch off hyperparameter optimization).
        iterations : int, optional
            Number of training iterations.
        topic_threshold : float, optional
            Threshold of the probability above which we consider a topic.
        random_seed: int, optional
            Random seed to ensure consistent results, if 0 - use system clock.

        """
        super().__init__(mallet_path, corpus=corpus, num_topics=num_topics, alpha=alpha, beta=beta, id2word=id2word,
                         workers=workers, prefix=prefix, optimize_interval=optimize_interval, iterations=iterations,
                         topic_threshold=topic_threshold, random_seed=random_seed, noise_words_max=noise_words_max,
                         skew=skew, embedding_path=embedding_path, closest_x_words=closest_x_words,
                         save_embedding_comp=save_embedding_comp, is_parent=True)
        self.alpha_array_infile = alpha_array_infile
        self.tw_dist_file = tw_dist_file
        self.noise_dist_file = noise_dist_file
        if corpus is not None:
            self.train(corpus)

    def ftwdistfile(self):
        '''
        Get path to noise file

        :return:
        ---------
        str
        |   Path to noise distribution file.
        '''
        return self.prefix + 'tw_dist.csv'

    def falphaarrayfile(self):
        '''
        Get path to noise file

        :return:
        ---------
        str
        |   Path to noise distribution file.
        '''
        return self.prefix + 'alpha_array.csv'

    def fbetafile(self):
        '''
        Get path to noise file

        :return:
        ---------
        str
        |   Path to noise distribution file.
        '''
        return self.prefix + 'beta.csv'

    def train(self, corpus):
        """Train Mallet LDA.

        Parameters
        ----------
        corpus : iterable of iterable of (int, int)
            Corpus in BoW format

        """
        self.get_closest_embedded_words()
        self.convert_input(corpus, infer=False)
        cmd = self.mallet_path + ' train-topics --input %s --num-topics %s  --alpha %s --optimize-interval %s '\
            '--num-threads %s --output-state %s --output-doc-topics %s --output-topic-keys %s '\
            '--num-iterations %s --inferencer-filename %s --doc-topics-threshold %s  --random-seed %s --output-noise %s '\
            '--noise-words-max %s --skew %s --beta %s --close-words-file %s --tau %s --optimize-burn-in %s ' \
            '--topic-word-weights-file %s --output-alpha-array %s --output-beta %s '

        cmd = cmd % (
            self.fcorpusmallet(), self.num_topics, self.alpha, self.optimize_interval,
            self.workers, self.fstate(), self.fdoctopics(), self.ftopickeys(), self.iterations,
            self.finferencer(), self.topic_threshold, str(self.random_seed), self.fnoisefile(), str(self.noise_words_max),
            str(self.skew), str(self.beta), self.fclosewords(), str(self.tau), str(self.tau),
            self.fwordweights(), self.falphaarrayfile(), self.fbetafile()
        )
        if self.alpha_array_infile is not None:
            cmd += '--alpha-array-infile %s ' % self.alpha_array_infile
        else:
            cmd += '--alpha %s ' % str(self.alpha)
        if self.tw_dist_file is not None:
            cmd += '--tw-dist-infile %s ' % self.tw_dist_file
        if self.noise_dist_file is not None:
            cmd += '--noise-dist-infile %s ' % self.noise_dist_file
        # NOTE "--keep-sequence-bigrams" / "--use-ngrams true" poorer results + runs out of memory
        logger.info("training MALLET LDA with %s", cmd)
        check_output(args=cmd, shell=True)
        self.word_topics = self.load_word_topics()
        # NOTE - we are still keeping the wordtopics variable to not break backward compatibility.
        # word_topics has replaced wordtopics throughout the code;
        # wordtopics just stores the values of word_topics when train is called.
        self.wordtopics = self.word_topics

    def load_alpha_array(self):
        alpha = []
        logger.info("loading alpha array from %s", self.falphaarrayfile())
        with open(self.falphaarrayfile(), 'r') as f:
            for line in f:
                prob = line.strip()
                alpha.append(float(prob))
        return alpha

    def load_beta(self):
        beta = 0
        logger.info("loading beta from %s", self.fbetafile())
        with open(self.fbetafile(), 'r') as f:
            for line in f:
                prob = line.strip()
                beta = float(prob)
        return beta



from .detnd import deTND
from .dtnd import dTND
from .gtm import GTM
from .dnlda import dNLDA
from .denlda import deNLDA
from .nlda import NLDA
from .enlda import eNLDA
from .tnd import TND
from .etnd import eTND



#!/home/rob/.env/topics/bin/python
from gensim.models.fasttext import load_facebook_vectors
from ..helpers.common import save_topics, save_noise_dist
from ..wrappers import deTNDMallet
from .dtnd import dTND


class deTND(dTND):
    topics = None
    last_alpha_array_file = None
    last_beta = None
    last_noise_dist_file = None
    last_tw_dist_file = None

    def __init__(self, dataset=None, k=30, alpha=50, beta0=0.01, beta1=25, noise_words_max=200,
                 iterations=1000, top_words=20, tw_dist=None, corpus=None, dictionary=None,
                 save_path=None, mallet_tnd_path=None, starting_alpha_array_file=None, embedding_path=None,
                 closest_x_words=3,
                 random_seed=1824, run=True,
                 workers=4):
        '''

        :param dataset: dataset should be an ordered list of sub-datasets,
                        where dataset[i] is the data set for time period i
        :param k:
        :param alpha:
        :param beta0:
        :param beta1:
        :param noise_words_max:
        :param iterations:
        :param top_words:
        :param tw_dist:
        :param corpus:
        :param dictionary:
        :param save_path:
        :param mallet_tnd_path:
        :param starting_beta_file:
        :param starting_alpha_array_file:
        :param random_seed:
        :param run:
        '''

        super().__init__(dataset, k, alpha, beta0, beta1, noise_words_max, iterations,
                         top_words, tw_dist, corpus, dictionary, save_path, mallet_tnd_path,
                         starting_alpha_array_file, random_seed, run=False, workers=workers)
        if save_path is not None:
            save = True
            self.save_path = save_path
        else:
            save = False
            self.save_path = 'detnd_results/'
        self.embedding_path = embedding_path
        self.closest_x_words = closest_x_words

        if run:
            self.run_all_time_periods(save=save)

    def run_one_time_period(self, t):
        # pass previous noise and tw distribution files into mallet
        # prepare data for time period t
        self.prepare_data(t)
        # run model
        if self.last_alpha_array_file is not None:
            model = deTNDMallet(self.mallet_path, corpus=self.corpus, num_topics=self.k, beta=self.last_beta,
                                id2word=self.dictionary, iterations=self.iterations, skew=self.beta1,
                                noise_words_max=self.noise_words_max, workers=self.workers,
                                noise_dist_file=self.last_noise_dist_file, tw_dist_file=self.last_tw_dist_file,
                                alpha_array_infile=self.last_alpha_array_file, embedding_path=self.embedding_path,
                                closest_x_words=self.closest_x_words)
        else:
            model = deTNDMallet(self.mallet_path, corpus=self.corpus, num_topics=self.k, beta=self.last_beta,
                                id2word=self.dictionary, iterations=self.iterations, skew=self.beta1,
                                noise_words_max=self.noise_words_max, workers=self.workers,
                                alpha=self.alpha, embedding_path=self.embedding_path,
                                closest_x_words=self.closest_x_words)
        # get/set new beta and alpha array files, noise dist file
        self.last_beta = model.load_beta()
        self.last_alpha_array_file = model.falphaarrayfile()
        self.last_noise_dist_file = model.fnoisefile()
        self.last_tw_dist_file = model.fwordweights()
        return model

    def run_all_time_periods(self, save=True):
        for t in range(0, len(self.dataset)):
            model = self.run_one_time_period(t)
            topics = model.show_topics(num_topics=self.k, num_words=self.top_words, formatted=False)
            noise = model.load_noise_dist()
            self.topics = topics
            self.noise_distribution = noise
            if save:
                topics = model.show_topics(num_topics=self.k, num_words=self.top_words, formatted=False)
                topics = [[w for (w, _) in topic[1]] for topic in topics]
                save_topics(topics, self.save_path + 'topics_{}_{}_{}.csv'.format(self.k, self.closest_x_words, t))
                noise_list = sorted([(x, noise[x]) for x in self.noise_distribution.keys()], key=lambda x: x[1], reverse=True)
                save_noise_dist(noise_list, self.save_path + 'noise_{}_{}_{}.csv'.format(self.k, self.closest_x_words, t))

    def __str__(self):
        return 'deTND'



#!/home/rob/.env/topics/bin/python
import math
import random
from gensim import corpora
from ..helpers.exceptions import MissingModelError, MissingDataSetError, MissingSeedsError
from ..helpers.common import load_flat_dataset
from ..helpers.weighting import compute_idf_weights
from ..wrappers import GTMMallet, TNDMallet


class GTM:
    topics = None

    def __init__(self, dataset=None, tnd_k=30, tnd_alpha=50, tnd_beta0=0.01, tnd_beta1=25, tnd_noise_words_max=200,
                 tnd_iterations=1000, gtm_iterations=1000, gtm_k=30, phi=10, topic_depth=100, top_words=20,
                 tnd_noise_distribution=None, gtm_tw_dist=None, gtm_topics=None, corpus=None, dictionary=None,
                 save_path=None, mallet_tnd_path=None, mallet_gtm_path=None, random_seed=1824, run=True,
                 tnd_workers=4, gtm_workers=4, seed_topics_file=None, over_sampling_factor=1, seed_gpu_weights=None):
        self.dataset = dataset
        self.tnd_k = tnd_k
        self.tnd_alpha = tnd_alpha
        self.tnd_beta0 = tnd_beta0
        self.tnd_beta1 = tnd_beta1
        self.tnd_noise_words_max = tnd_noise_words_max
        self.tnd_iterations = tnd_iterations
        self.gtm_iterations = gtm_iterations
        self.gtm_k = gtm_k
        self.phi = phi
        self.topic_depth = topic_depth
        self.top_words = top_words
        self.tnd_noise_distribution = tnd_noise_distribution
        self.gtm_tw_dist = gtm_tw_dist
        self.gtm_topics = gtm_topics
        self.corpus = corpus
        self.dictionary = dictionary
        if save_path is not None:
            self.save_path = save_path
        else:
            self.save_path = 'gtm_results/'
        self.mallet_tnd_path = mallet_tnd_path
        self.mallet_gtm_path = mallet_gtm_path
        self.random_seed = random_seed
        random.seed(self.random_seed)
        self.gtm_workers = gtm_workers
        self.tnd_workers = tnd_workers
        self.seed_topics_file = seed_topics_file
        self.seed_gpu_weights = seed_gpu_weights
        self.over_sampling_factor = over_sampling_factor

        if self.seed_topics_file is None:
            raise MissingSeedsError

        if self.mallet_tnd_path is None and self.tnd_noise_distribution is None:
            raise MissingModelError('tnd')
        if self.mallet_gtm_path is None and self.gtm_tw_dist is None:
            raise MissingModelError('gtm')
        if self.dataset is None and (self.corpus is None or self.dictionary is None
                                     or self.tnd_noise_distribution is None or self.gtm_tw_dist is None):
            raise MissingDataSetError

        if self.seed_gpu_weights is None and self.dataset is not None:
            seed_topic_words = load_flat_dataset(seed_topics_file, delimiter=',')
            self.seed_gpu_weights = compute_idf_weights(dataset, seed_topic_words)
        elif self.seed_gpu_weights is None:
            raise MissingDataSetError('You must input either a data set to compute seed gpu weights, or precomputed seed gpu weights.')

        if run:
            if (self.dataset is not None) and (self.corpus is None or self.dictionary is None):
                self.prepare_data()
            if self.tnd_noise_distribution is None:
                self.compute_tnd()
            if self.gtm_tw_dist is None:
                self.compute_gtm()

            self.filter_noise()

    def prepare_data(self):
        """
        takes dataset, sets self.dictionary and self.corpus for use in Mallet models
        :return: void
        """
        dictionary = corpora.Dictionary(self.dataset)
        dictionary.filter_extremes()
        corpus = [dictionary.doc2bow(doc) for doc in self.dataset]
        self.dictionary = dictionary
        self.corpus = corpus

    def compute_tnd(self):
        """
        takes dataset, tnd parameters, tnd mallet path, and computes tnd model on dataset
        sets self.tnd_noise_distribution to the noise distribution computed in tnd
        :return: void
        """
        model = TNDMallet(self.mallet_tnd_path, self.corpus, num_topics=self.tnd_k, id2word=self.dictionary,
                          workers=self.tnd_workers,
                          alpha=self.tnd_alpha, beta=self.tnd_beta0, skew=self.tnd_beta1,
                          iterations=self.tnd_iterations, noise_words_max=self.tnd_noise_words_max,
                          random_seed=self.random_seed)
        noise = model.load_noise_dist()
        self.tnd_noise_distribution = noise

    def compute_gtm(self):
        """
        takes dataset, gtm parameters, gtm mallet path, and computes gtm model on dataset
        sets self.gtm_tw_dist to the topic word distribution computed in gtm
        :return: void
        """
        model = GTMMallet(self.mallet_gtm_path, self.corpus, num_topics=self.gtm_k, id2word=self.dictionary,
                          workers=self.gtm_workers,
                          iterations=self.gtm_iterations, random_seed=self.random_seed,
                          seed_topics_file=self.seed_topics_file, over_sampling_factor=self.over_sampling_factor,
                          seed_gpu_weights=self.seed_gpu_weights)
        topic_word_distribution = model.load_word_topics()
        self.gtm_tw_dist = topic_word_distribution
        topics = model.show_topics(num_topics=self.gtm_k, num_words=self.topic_depth, formatted=False)
        self.gtm_topics = [[w for (w, _) in topics[i][1]] for i in range(0, len(topics))]

    def filter_noise(self):
        """
        takes self.tnd_noise_distribution, self.gtm_tw_dist, self.phi, self.top_words, and computes Ngtm topics
        sets self.topics to the set of topics computed from noise distribution and topic word distribution
        :return: void
        """
        topics = []
        for i in range(0, len(self.gtm_topics)):
            topic = self.gtm_topics[i]
            final_topic = []
            j = 0
            while len(final_topic) < self.top_words and j < len(topic) and j < self.topic_depth:
                w = topic[j]
                token_id = self.dictionary.token2id[w]
                beta = 2
                if w in self.tnd_noise_distribution:
                    beta += self.tnd_noise_distribution[w]
                beta = max(2, beta * (self.phi / self.gtm_k))
                alpha = 2 + self.gtm_tw_dist[i, token_id]
                roll = random.betavariate(alpha=math.sqrt(alpha), beta=math.sqrt(beta))
                if roll >= 0.5:
                    final_topic.append(w)
                    if w not in self.tnd_noise_distribution:
                        self.tnd_noise_distribution[w] = 0
                    self.tnd_noise_distribution[w] += (alpha - 2)
                j += 1
            topics.append(final_topic)
        self.topics = topics

    def get_topics(self, top_words=None):
        """
        takes top_words and self.topics, returns a list of topic lists of length top_words
        :param top_words:
        :return: list of topic lists
        """
        if top_words is None:
            top_words = self.top_words
        topics = self.topics
        if topics is None or len(topics) < 1:
            raise ValueError('No topics have been computed yet.')

        return [x[:top_words] for x in topics]

    def get_noise_distribution(self, tnd_noise_words_max=None):
        """
        takes self.tnd_noise_distribution and tnd_noise_words_max
        returns a list of (noise word, frequency) tuples ranked by frequency
        :param tnd_noise_words_max:
        :return: list of (noise word, frequency) tuples
        """
        if tnd_noise_words_max is None:
            tnd_noise_words_max = self.tnd_noise_words_max
        noise = self.tnd_noise_distribution
        if noise is None or len(noise) < 1:
            raise ValueError('No noise distribution has been computed yet.')

        noise_list = sorted([(x, int(noise[x])) for x in noise.keys()], key=lambda x: x[1], reverse=True)
        return noise_list[:tnd_noise_words_max]



'''
Exceptions shared across files.
'''

class MissingModelError(Exception):
    def __init__(self, model, message=None):
        self.model = model
        self.message = message
        if self.message is None:
            self.message = 'You must either define the path to the {} model, or provide the pretrained model.'
        super().__init__(self.message.format(model))


class MissingDataSetError(Exception):
    def __init__(self, message=None):
        self.message = message
        if self.message is None:
            self.message = 'You must either provide a pretrained noise distribution and pretrained topic-word ' \
                           'distribution, or provide a data set to be modeled.'
        super().__init__(self.message)


class MissingSeedsError(Exception):
    def __init__(self, message=None):
        self.message = message
        if self.message is None:
            self.message = 'You must provide a seed topics file.'
        super().__init__(self.message)


class MissingSeedWeightsError(Exception):
    def __init__(self, message=None):
        self.message = message
        if self.message is None:
            self.message = 'You must provide a seed word weighting map.  ' \
                           'You can do this easily using: helpers.weighting.compute_idf_weights'
        super().__init__(self.message)


class MissingEmbeddingPathError(Exception):
    def __init__(self, message=None):
        self.message = message
        if self.message is None:
            self.message = 'You must provide a path to a pre-trained embedding space.'
        super().__init__(self.message)



#!/home/rob/.env/topics/bin/python
from gensim import corpora
from ..helpers.common import save_topics, save_noise_dist
from ..wrappers import dTNDMallet
from .tnd import TND


class dTND(TND):
    topics = None
    last_alpha_array_file = None
    last_beta = None
    last_noise_dist_file = None
    last_tw_dist_file = None
    noise_distribution = []

    def __init__(self, dataset=None, k=30, alpha=50, beta0=0.01, beta1=25, noise_words_max=200,
                 iterations=1000, top_words=20, topic_word_distribution=None, corpus=None, dictionary=None,
                 save_path=None, mallet_path=None, starting_alpha_array_file=None, random_seed=1824, run=True,
                 noise_distribution=None, workers=4):
        '''

        :param dataset: dataset should be an ordered list of sub-datasets,
                        where dataset[i] is the data set for time period i
        :param k:
        :param alpha:
        :param beta0:
        :param beta1:
        :param noise_words_max:
        :param iterations:
        :param top_words:
        :param noise_distribution:
        :param tw_dist:
        :param corpus:
        :param dictionary:
        :param save_path:
        :param mallet_tnd_path:
        :param starting_beta_file:
        :param starting_alpha_array_file:
        :param random_seed:
        :param run:
        '''

        super().__init__(dataset=dataset, k=k, alpha=alpha, beta0=beta0, beta1=beta1, noise_words_max=noise_words_max,
                         iterations=iterations, top_words=top_words, topic_word_distribution=topic_word_distribution,
                         noise_distribution=noise_distribution, corpus=corpus, dictionary=dictionary,
                         save_path=save_path, mallet_path=mallet_path, random_seed=random_seed, run=False,
                         workers=workers)
        if save_path is not None:
            save = True
            self.save_path = save_path
        else:
            save = False
            self.save_path = 'dtnd_results/'
        self.last_beta = self.beta0
        self.last_alpha_array_file = starting_alpha_array_file

        if run:
            self.run_all_time_periods(save=save)

    def prepare_data(self, t):
        """
        takes dataset, sets self.dictionary and self.corpus for use in Mallet models and NLDA
        :return: void
        """
        dictionary = corpora.Dictionary(self.dataset[t])
        dictionary.filter_extremes()
        corpus = [dictionary.doc2bow(doc) for doc in self.dataset[t]]
        self.dictionary = dictionary
        self.corpus = corpus

    def run_one_time_period(self, t):
        # pass previous noise and tw distribution files into mallet
        # prepare data for time period t
        self.prepare_data(t)
        # run model
        model = dTNDMallet(self.mallet_path, corpus=self.corpus, num_topics=self.k, beta=self.last_beta,
                           id2word=self.dictionary, iterations=self.iterations, skew=self.beta1,
                           noise_words_max=self.noise_words_max, workers=self.workers,
                           noise_dist_file=self.last_noise_dist_file, tw_dist_file=self.last_tw_dist_file,
                           alpha_array_infile=self.last_alpha_array_file, alpha=self.alpha)
        # get/set new beta and alpha array files, noise dist file
        self.last_beta = model.load_beta()
        self.last_alpha_array_file = model.falphaarrayfile()
        self.last_noise_dist_file = model.fnoisefile()
        self.last_tw_dist_file = model.fwordweights()
        return model

    def run_all_time_periods(self, save=True):
        for t in range(0, len(self.dataset)):
            model = self.run_one_time_period(t)
            topics = model.show_topics(num_topics=self.k, num_words=self.top_words, formatted=False)
            noise = model.load_noise_dist()
            self.topics = topics
            self.noise_distribution.append(noise)
            if save:
                topics = model.show_topics(num_topics=self.k, num_words=self.top_words, formatted=False)
                topics = [[w for (w, _) in topic[1]] for topic in topics]
                save_topics(topics, self.save_path + 'topics_{}_{}.csv'.format(self.k, t))
                noise_list = sorted([(x, noise[x]) for x in noise.keys()], key=lambda x: x[1], reverse=True)
                save_noise_dist(noise_list, self.save_path + 'noise_{}_{}.csv'.format(self.k, t))

    def get_topics(self, t, top_words=None):
        """
        takes top_words and self.topics, returns a list of topic lists of length top_words
        :param t: time period
        :param top_words:
        :return: list of topic lists
        """
        if top_words is None:
            top_words = self.top_words
        topics = self.topics
        if len(topics) < 1:
            raise ValueError('No topics have been computed yet.')
        elif len(topics[t]) < 1:
            raise ValueError('No topics have been computed for this time period yet.')

        return [x[:top_words] for x in topics[t]]

    def get_noise_distribution(self, t, noise_words_max=None):
        """
        takes self.tnd_noise_distribution and tnd_noise_words_max
        returns a list of (noise word, frequency) tuples ranked by frequency
        :param t: time period
        :param tnd_noise_words_max:
        :return: list of (noise word, frequency) tuples
        """
        if noise_words_max is None:
            noise_words_max = self.noise_words_max
        if len(self.noise_distribution) < 1:
            raise ValueError('No noise distribution has been computed yet.')
        elif len(self.noise_distribution[t]) < 1:
            raise ValueError('No noise distribution has been computed for this time period yet.')

        noise = self.noise_distribution[t]
        noise_list = sorted([(x, int(noise[x])) for x in noise.keys()], key=lambda x: x[1], reverse=True)
        return noise_list[:noise_words_max]

    def __str__(self):
        return 'dTND'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 Radim Rehurek <radimrehurek@seznam.cz>
# Licensed under the GNU LGPL v2.1 - http://www.gnu.org/licenses/lgpl.html
# Refactored to run TND and save noise distribution by Rob Churchill (Feb 2021)

r"""Python wrapper for `Topic Noise Discriminator (TND)`_
adapted from `MALLET, the Java topic modelling toolkit <http://mallet.cs.umass.edu/>`_

This module allows both TND model estimation from a training corpus and inference of topic distribution on new,
unseen documents, using an (optimized version of) collapsed gibbs sampling from MALLET.

Notes
-----
MALLET's LDA training requires :math:`O(corpus\_words)` of memory, keeping the entire corpus in RAM.
If you find yourself running out of memory, either decrease the `workers` constructor parameter,
or use :class:`gensim.models.ldamodel.LdaModel` or :class:`gensim.models.ldamulticore.LdaMulticore`
which needs only :math:`O(1)` memory.
The wrapped model can NOT be updated with new documents for online training -- use
:class:`~gensim.models.ldamodel.LdaModel` or :class:`~gensim.models.ldamulticore.LdaMulticore` for that.

"""


import logging
import os

from gensim.utils import check_output
from gensim.models.fasttext import load_facebook_vectors
from .base_wrapper import BaseMalletWrapper

logger = logging.getLogger(__name__)


class TNDMallet(BaseMalletWrapper):
    """Python wrapper for TND using `MALLET <http://mallet.cs.umass.edu/>`_.

    Communication between MALLET and Python takes place by passing around data files on disk
    and calling Java with subprocess.call().

    Warnings
    --------
    This is **only** python wrapper for `MALLET TND`_,
    you need to install original implementation first and pass the path to binary to ``mallet_path``.

    """
    def __init__(self, mallet_path, corpus=None, num_topics=100, alpha=50, beta=0.01, id2word=None, workers=4, prefix=None,
                 optimize_interval=0, iterations=1000, topic_threshold=0.0, random_seed=0, noise_words_max=100, skew=0,
                 is_parent=False):
        """

        Parameters
        ----------
        mallet_path : str
            Path to the mallet binary, e.g. `/home/username/mallet-2.0.7/bin/mallet`.
        corpus : iterable of iterable of (int, int), optional
            Collection of texts in BoW format.
        num_topics : int, optional
            Number of topics.
        alpha : int, optional
            Alpha parameter of LDA.
        id2word : :class:`~gensim.corpora.dictionary.Dictionary`, optional
            Mapping between tokens ids and words from corpus, if not specified - will be inferred from `corpus`.
        workers : int, optional
            Number of threads that will be used for training.
        prefix : str, optional
            Prefix for produced temporary files.
        optimize_interval : int, optional
            Optimize hyperparameters every `optimize_interval` iterations
            (sometimes leads to Java exception 0 to switch off hyperparameter optimization).
        iterations : int, optional
            Number of training iterations.
        topic_threshold : float, optional
            Threshold of the probability above which we consider a topic.
        random_seed: int, optional
            Random seed to ensure consistent results, if 0 - use system clock.

        """
        super().__init__(mallet_path, corpus=corpus, num_topics=num_topics, alpha=alpha, beta=beta, id2word=id2word,
                         workers=workers, prefix=prefix, optimize_interval=optimize_interval, iterations=iterations,
                         topic_threshold=topic_threshold, random_seed=random_seed)
        self.noise_words_max = noise_words_max
        self.skew = skew
        if corpus is not None and not is_parent:
            self.train(corpus)


    def fnoisefile(self):
        '''
        Get path to noise file

        :return:
        ---------
        str
        |   Path to noise distribution file.
        '''
        return self.prefix + 'noise_dist.csv'

    def train(self, corpus):
        """Train Mallet LDA.

        Parameters
        ----------
        corpus : iterable of iterable of (int, int)
            Corpus in BoW format

        """
        self.convert_input(corpus, infer=False)
        cmd = self.mallet_path + ' train-topics --input %s --num-topics %s  --alpha %s --optimize-interval %s '\
            '--num-threads %s --output-state %s --output-doc-topics %s --output-topic-keys %s '\
            '--num-iterations %s --inferencer-filename %s --doc-topics-threshold %s  --random-seed %s --output-noise %s '\
            '--noise-words-max %s --skew %s --beta %s'

        cmd = cmd % (
            self.fcorpusmallet(), self.num_topics, self.alpha, self.optimize_interval,
            self.workers, self.fstate(), self.fdoctopics(), self.ftopickeys(), self.iterations,
            self.finferencer(), self.topic_threshold, str(self.random_seed), self.fnoisefile(), str(self.noise_words_max),
            str(self.skew), str(self.beta)
        )
        # NOTE "--keep-sequence-bigrams" / "--use-ngrams true" poorer results + runs out of memory
        logger.info("training MALLET LDA with %s", cmd)
        check_output(args=cmd, shell=True)
        self.word_topics = self.load_word_topics()
        # NOTE - we are still keeping the wordtopics variable to not break backward compatibility.
        # word_topics has replaced wordtopics throughout the code;
        # wordtopics just stores the values of word_topics when train is called.
        self.wordtopics = self.word_topics

    def load_noise_dist(self):
        noise_dist = {}
        logger.info("loading assigned topics from %s", self.fnoisefile())
        with open(self.fnoisefile(), 'r') as f:
            for line in f:
                word, val = line.strip().split(', ')
                noise_dist[word] = int(val)
        return noise_dist


r"""Python wrapper for `embedded TND `_
from `MALLET, the Java topic modelling toolkit <http://mallet.cs.umass.edu/>`_

This module allows both eTND model estimation from a training corpus and inference of topic distribution on new,
unseen documents, using an (optimized version of) collapsed gibbs sampling from MALLET.

Notes
-----
MALLET's LDA training requires :math:`O(corpus\_words)` of memory, keeping the entire corpus in RAM.
If you find yourself running out of memory, either decrease the `workers` constructor parameter,
or use :class:`gensim.models.ldamodel.LdaModel` or :class:`gensim.models.ldamulticore.LdaMulticore`
which needs only :math:`O(1)` memory.
The wrapped model can NOT be updated with new documents for online training -- use
:class:`~gensim.models.ldamodel.LdaModel` or :class:`~gensim.models.ldamulticore.LdaMulticore` for that.

Installation
------------
Use `official guide <http://mallet.cs.umass.edu/download.php>`_ or this one ::

    sudo apt-get install default-jdk
    sudo apt-get install ant
    git clone git@github.com:mimno/Mallet.git
    cd Mallet/
    ant

Examples
--------
.. sourcecode:: pycon

    >>> from gensim.test.utils import common_corpus, common_dictionary
    >>> from gensim.models.wrappers import LdaMallet
    >>>
    >>> path_to_mallet_binary = "/path/to/mallet/binary"
    >>> model = LdaMallet(path_to_mallet_binary, corpus=common_corpus, num_topics=20, id2word=common_dictionary)
    >>> vector = model[common_corpus[0]]  # LDA topics of a documents

"""


class eTNDMallet(TNDMallet):
    """Python wrapper for eTND using `MALLET <http://mallet.cs.umass.edu/>`_.

    Communication between MALLET and Python takes place by passing around data files on disk
    and calling Java with subprocess.call().

    Warnings
    --------
    This is **only** python wrapper for `MALLET LDA <http://mallet.cs.umass.edu/>`_,
    you need to install original implementation first and pass the path to binary to ``mallet_path``.

    """
    def __init__(self, mallet_path, corpus=None, num_topics=100, alpha=50, beta=0.01, id2word=None, workers=4, prefix=None,
                 optimize_interval=0, iterations=1000, topic_threshold=0.0, random_seed=0, noise_words_max=100, skew=0,
                 tau=200, embedding_path=None, closest_x_words=3, save_embedding_comp=True, is_parent=False):
        """

        Parameters
        ----------
        mallet_path : str
            Path to the mallet binary, e.g. `/home/username/mallet-2.0.7/bin/mallet`.
        corpus : iterable of iterable of (int, int), optional
            Collection of texts in BoW format.
        num_topics : int, optional
            Number of topics.
        alpha : int, optional
            Alpha parameter of LDA.
        id2word : :class:`~gensim.corpora.dictionary.Dictionary`, optional
            Mapping between tokens ids and words from corpus, if not specified - will be inferred from `corpus`.
        workers : int, optional
            Number of threads that will be used for training.
        prefix : str, optional
            Prefix for produced temporary files.
        optimize_interval : int, optional
            Optimize hyperparameters every `optimize_interval` iterations
            (sometimes leads to Java exception 0 to switch off hyperparameter optimization).
        iterations : int, optional
            Number of training iterations.
        topic_threshold : float, optional
            Threshold of the probability above which we consider a topic.
        random_seed: int, optional
            Random seed to ensure consistent results, if 0 - use system clock.

        """
        super().__init__(mallet_path, corpus=corpus, num_topics=num_topics, alpha=alpha, beta=beta, id2word=id2word,
                         workers=workers, prefix=prefix, optimize_interval=optimize_interval, iterations=iterations,
                         topic_threshold=topic_threshold, random_seed=random_seed, noise_words_max=noise_words_max,
                         skew=skew, is_parent=True)
        self.tau = tau
        self.embedding_path = embedding_path
        self.closest_x_words = closest_x_words
        self.save_embedding_comp = save_embedding_comp
        if corpus is not None and not is_parent:
            self.train(corpus)

    def fclosewords(self):
        """Get path to closest words to each word in embedding space

        Returns
        -------
        str
        """
        # if os.path.exists('data/cw_{}_{}.txt'.format(self.embedding_path.replace('.bin', ''), self.closest_x_words)):
        #     return 'data/cw_{}'.format(self.embedding_path.replace('.bin', '.txt'))
        return self.prefix + 'closewords.txt'

    def get_closest_embedded_words(self):
        if os.path.exists('data/cw_{}_{}.txt'.format(self.embedding_path.replace('.bin', ''), self.closest_x_words)):
            return
        vec = load_facebook_vectors(self.embedding_path)
        vec.fill_norms()
        hasht = dict()
        for w in vec.key_to_index:
            hasht[w] = [x[0] for x in vec.most_similar_cosmul(w, topn=self.closest_x_words)]
        with open(self.fclosewords(), 'w') as f:
            for w in hasht.keys():
                f.write('{},{}\n'.format(w, ','.join(hasht[w])))
        # if self.save_embedding_comp:
        #     with open('data/cw_{}_{}.txt'.format(self.embedding_path.replace('.bin', ''), self.closest_x_words, 'w')) as f:
        #         for w in hasht.keys():
        #             f.write('{},{}\n'.format(w, ','.join(hasht[w])))

    def train(self, corpus):
        """Train Mallet LDA.

        Parameters
        ----------
        corpus : iterable of iterable of (int, int)
            Corpus in BoW format

        """
        self.get_closest_embedded_words()
        self.convert_input(corpus, infer=False)
        cmd = self.mallet_path + ' train-topics --input %s --num-topics %s  --alpha %s --optimize-interval %s '\
            '--num-threads %s --output-state %s --output-doc-topics %s --output-topic-keys %s '\
            '--num-iterations %s --inferencer-filename %s --doc-topics-threshold %s  --random-seed %s --output-noise %s '\
            '--noise-words-max %s --skew %s --beta %s --close-words-file %s --tau %s --optimize-burn-in %s'

        cmd = cmd % (
            self.fcorpusmallet(), self.num_topics, self.alpha, self.optimize_interval,
            self.workers, self.fstate(), self.fdoctopics(), self.ftopickeys(), self.iterations,
            self.finferencer(), self.topic_threshold, str(self.random_seed), self.fnoisefile(), str(self.noise_words_max),
            str(self.skew), str(self.beta), self.fclosewords(), str(self.tau), str(self.tau)
        )
        # NOTE "--keep-sequence-bigrams" / "--use-ngrams true" poorer results + runs out of memory
        logger.info("training MALLET LDA with %s", cmd)
        check_output(args=cmd, shell=True)
        self.word_topics = self.load_word_topics()
        # NOTE - we are still keeping the wordtopics variable to not break backward compatibility.
        # word_topics has replaced wordtopics throughout the code;
        # wordtopics just stores the values of word_topics when train is called.
        self.wordtopics = self.word_topics
